#include "sysincludes.h"

/* Raise the number of currently open files in the process. */

/* If NEW_LIMIT == -1 set the limit to the maximum allowable.
   Otherwise, set it to be the value of NEW_LIMIT. 

   Note the insanity that the maximum limit appears
   to be 200 open files, but NOFILE and FD_SETSIZE are
   both set to 256!!! */

int 
raise_file_limit (int new_limit)
{
  struct rlimit rl;

#if defined (RLIMIT_NOFILE)
  if (::getrlimit (RLIMIT_NOFILE, &rl) == -1)
    return -1;
#else
  rl.rlim_cur = NOFILES_MAX;
#endif /* RLIMIT_NOFILE */

  if (new_limit < 0 || new_limit > rl.rlim_max)
    rl.rlim_cur = rl.rlim_max;
  else
    rl.rlim_cur = new_limit;

  return ::setrlimit (RLIMIT_NOFILE, &rl);
}

/* flags are file status flags to turn on */

int
set_fl (int fd, int flags)
{
  int val;

  if ((val = ::fcntl (fd, F_GETFL, 0)) == -1)
    return -1;

  val |= flags; /* turn on flags */

  if (::fcntl (fd, F_SETFL, val) == -1)
    return -1;
  else
    return 0;
}

/* flags are the file status flags to turn off */

int
clr_fl (int fd, int flags)
{
  int val;

  if ((val = ::fcntl (fd, F_GETFL, 0)) == -1)
    return -1;

  val &= ~flags; /* turn flags off */

  if (::fcntl (fd, F_SETFL, val) == -1)
    return -1;
  else
    return 0;
}

/* The following code is courtesy of Rich Salz and may
   be useful for odd systems that lack writev(). */

#if defined (NEED_WRITEV)
/*  $Revision: 1.3 $
**
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>

/* Hacks since you don't have INN's header file. */
#define POINTER	char*
#define SIZE_T	int

/*
**  Keep writing until everything has been written or we get an error.
*/
int
xwrite(int fd, char *p, int i)
{
  int c;

  for ( ; i; p += c, i -= c)
    if ((c = ::write (fd, (POINTER) p, (SIZE_T) i)) <= 0)
      return -1;
  return 0;
}


/*
**  "Fake" writev for sites without it.
*/
int
writev (int fd, struct iovec *vp, int vpcount)
{
  int count;

  for (count = 0; --vpcount >= 0; count += vp->iov_len, vp++)
    if (xwrite (fd, vp->iov_base, vp->iov_len) < 0)
      return -1;
  return count;
}
#endif // (NEED_WRITEV)
